// ------------------------------------------------------------------------
// audioio-ewf.cpp: Ecasound wave format input/output
// Copyright (C) 1999 Kai Vehmanen (kaiv@wakkanet.fi)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
// ------------------------------------------------------------------------

#include <algorithm>
#include <string>
#include <iostream.h>
#include <fstream.h>
#include <cmath>

#include "samplebuffer.h"
#include "audioio-ewf.h"

#include "error.h"
#include "debug.h"
  
EWFFILE::EWFFILE (const string& name, const SIMODE mode, const AIO_PARAMS& fmt) 
  :  AUDIO_IO_DEVICE(name, mode, fmt) {
    //
    // Prepare data file (.wav) for processing.
    //

    string::const_iterator e = find(name.begin(), name.end(), '.');

    if (e == name.end()) {
        throw(new ECA_ERROR("AUDIOIO-EWF", "Invalid file name; unable to open file.",stop));
        return;
    }
    wavename = string(name.begin(), e);
    wavename = wavename + ".wav";

    ecadebug->msg(1, "AUDIOIO-EWF: Opening EWF data file" + wavename);

    wobject = new WAVEFILE(wavename, mode, fmt);
    
    if (mode == si_read) {
        //
        // Read info from .ewf file.
        //

        read_ewf_parametres();
    }
    else
        sample_offset = 0;
        
    //
    // Initialize attributes.
    //
    curpos_in_samples = 0;
    wave_object_active = false;
}

EWFFILE::~EWFFILE(void) {
    write_ewf_parametres();
    delete wobject;    
}

void EWFFILE::get_sample(SAMPLE_BUFFER* t) {
    curpos_in_samples += t->length_in_samples();
    if (wave_object_active == false && position_in_samples() >= sample_offset) {
        ecadebug->msg(2, "AUDIOIO-EWF: wave_object activated" + ewfname);
        wave_object_active = true;
        wobject->position_in_samples(position_in_samples()
                                     - sample_offset);
    }
    if (wave_object_active == true)
        wobject->get_sample(t);
    else
        t->make_silent();
}

void EWFFILE::put_sample(SAMPLE_BUFFER* t) {
    if (wave_object_active == false) {
//        if (t->average_volume() == SAMPLE_BUFFER::silent_value) {
        if (t->average_volume() / SAMPLE_BUFFER::max_amplitude < 1.0 / 32767.0) {
            // --
            // Bufferin keskivolume on pienempi, mit 16bit sample-arvolla
            // voidaan esitt.
            // --
//            ecadebug->msg(15, "AUDIOIO-EWF:: silent t->average_volume()");
            curpos_in_samples += t->length_in_samples();
            return;
        }
        else {
            wave_object_active = true;
            sample_offset = curpos_in_samples;
	    MESSAGE_ITEM m;
	    m << "AUDIOIO-EWF: found sample_offset " << sample_offset << ".";
            ecadebug->msg(5, m.to_string());
        }
    }
//    ecadebug->msg(15, "AUDIOIO-EWF.cpp: value of t->average_volume()", t->average_volume());
    curpos_in_samples += t->length_in_samples();
    wobject->put_sample(t);
}

void EWFFILE::first(void) {
    curpos_in_samples = 0;
    wobject->first();
}

void EWFFILE::last(void) {
    if (wave_object_active == true) {
        wobject->last();
        curpos_in_samples = sample_offset + wobject->position_in_samples();
    }
    else
        curpos_in_samples = sample_offset;
}

long EWFFILE::length_in_samples(void)  const {
    return(sample_offset + wobject->length_in_samples());
}

long EWFFILE::position_in_samples(void)  const {
    return(curpos_in_samples);
}

void EWFFILE::position_in_samples(long pos) {
  //    finished(false);
    if (pos >= sample_offset) {
        wobject->position_in_samples(pos - sample_offset);
        curpos_in_samples = sample_offset + wobject->position_in_samples();
    }
    else {
        wave_object_active = false;
        curpos_in_samples = pos;
        wobject->first();
    }
}

void EWFFILE::read_ewf_parametres(void) {
    ifstream fin(ewfname.c_str());
    char c;
    while (fin.get(c)) {
        if (c == '=') break;
        ecadebug->msg(5, "AUDIOIO-EWF: r_ewf_params(), found a = sign");
    }
    fin >> sample_offset;
    //    ecadebug->msg(5, "AUDIOIO-EWF: r_ewf_params(), read a sample_offset of", sample_offset);
}

void EWFFILE::write_ewf_parametres(void) {
    string temp;
    ofstream fout(ewfname.c_str());
    fout << "sample_offset = ";
    fout << sample_offset;
    //    ecadebug->msg(4, "AUDIOIO-EWF.cpp: writing sample_offset", sample_offset);
}
